from typing import Sequence

from ..output.fields import permission_fields
from ..output.types import FieldSpec
from ..session import api_session
from ..utils import dedent as _d
from .base import BaseFunction, api_function

__all__ = ("Permission",)

_default_list_fields = (permission_fields["vfolder_host_permission_list"],)


class Permission(BaseFunction):
    @api_function
    @classmethod
    async def list(
        cls,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[str]:
        """
        Fetches the list of atomic permissions.

        :param fields: Additional permission query fields to fetch.
        """
        query = _d("""
            query {
                vfolder_host_permissions { $fields }
            }
        """)
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(query)
        return data["vfolder_host_permissions"]



from collections.abc import Sequence
from typing import Optional

from ..output.fields import user_resource_policy_fields
from ..output.types import FieldSpec
from ..session import api_session
from ..utils import dedent as _d
from .base import BaseFunction, api_function

_default_list_fields = (
    user_resource_policy_fields["name"],
    user_resource_policy_fields["max_vfolder_count"],
    user_resource_policy_fields["max_quota_scope_size"],
    user_resource_policy_fields["max_session_count_per_model_session"],
    user_resource_policy_fields["max_customized_image_count"],
)

_default_detail_fields = (
    user_resource_policy_fields["name"],
    user_resource_policy_fields["max_vfolder_count"],
    user_resource_policy_fields["max_quota_scope_size"],
    user_resource_policy_fields["max_session_count_per_model_session"],
    user_resource_policy_fields["max_customized_image_count"],
)


class UserResourcePolicy(BaseFunction):
    """
    Provides interactions with user resource policy.
    """

    _name: Optional[str]

    def __init__(self, name: Optional[str]) -> None:
        super().__init__()
        self._name = name

    @api_function
    @classmethod
    async def create(
        cls,
        name: str,
        *,
        max_vfolder_count: int,
        max_quota_scope_size: int,
        max_session_count_per_model_session: int,
        max_customized_image_count: int,
    ) -> dict:
        """
        Create a new user resource policy.

        :param name: Name of the user resource policy.
        :param max_vfolder_count: Maximum number of vfolders allowed.
        :param max_quota_scope_size: Maximum size of the quota scope (-1 for unlimited).
        :param max_session_count_per_model_session: Maximum session count per model service.
        :param max_customized_image_count: Maximum number of customized images allowed.
        :return: The created user resource policy data.
        """
        q = _d("""
            mutation($name: String!, $input: CreateUserResourcePolicyInput!) {
                create_user_resource_policy(name: $name, props: $input) {
                        ok
                        msg
                    }
            }
        """)

        input = {
            "max_vfolder_count": max_vfolder_count,
            "max_quota_scope_size": max_quota_scope_size,
            "max_session_count_per_model_session": max_session_count_per_model_session,
            "max_customized_image_count": max_customized_image_count,
        }

        result = await api_session.get().Admin._query(
            q,
            variables={
                "name": name,
                "input": input,
            },
        )

        return result["create_user_resource_policy"]

    @api_function
    @classmethod
    async def update(
        cls,
        name: str,
        *,
        max_vfolder_count: Optional[int] = None,
        max_quota_scope_size: Optional[int] = None,
        max_session_count_per_model_session: Optional[int] = None,
        max_customized_image_count: Optional[int] = None,
    ) -> dict:
        """
        Update an existing user resource policy with the given options.

        :param name: Name of the user resource policy to update.
        :param max_vfolder_count: New maximum number of vfolders allowed.
        :param max_quota_scope_size: New maximum size of the quota scope (-1 for unlimited).
        :param max_session_count_per_model_session: New maximum session count per model service.
        :param max_customized_image_count: New maximum number of customized images allowed.
        :return: The updated user resource policy data.
        """
        q = _d("""
            mutation($name: String!, $input: ModifyUserResourcePolicyInput!) {
                modify_user_resource_policy(name: $name, props: $input) {
                    ok
                }
            }
        """)

        input = {
            "max_vfolder_count": max_vfolder_count,
            "max_quota_scope_size": max_quota_scope_size,
            "max_session_count_per_model_session": max_session_count_per_model_session,
            "max_customized_image_count": max_customized_image_count,
        }

        variables = {
            "name": name,
            "input": {k: v for k, v in input.items() if v is not None},
        }

        result = await api_session.get().Admin._query(q, variables)
        return result["modify_user_resource_policy"]

    @api_function
    async def delete(self) -> dict:
        """
        Delete an existing user resource policy.
        :return: Result of the deletion operation.
        """
        q = _d("""
            mutation($name: String!) {
                delete_user_resource_policy(name: $name) {
                    ok
                    msg
                }
            }
        """)

        result = await api_session.get().Admin._query(q, {"name": self._name})
        return result["delete_user_resource_policy"]

    @api_function
    @classmethod
    async def list(
        cls,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Lists all user resource policies.

        :param fields: Fields to include in the output.
        :return: List of user resource policies.
        """
        q = "query { user_resource_policies { $fields } }"
        q = q.replace("$fields", " ".join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(q)
        return data["user_resource_policies"]

    @api_function
    async def get_info(
        self,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        """
        Get information about a specific user resource policy.

        :param name: Name of the user resource policy to retrieve.
        :param fields: Fields to include in the output.
        :return: The user resource policy data.
        """
        q = "query($name: String!) { user_resource_policy(name: $name) { $fields } }"
        q = q.replace("$fields", " ".join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(q, {"name": self._name})
        return data["user_resource_policy"]



import sys
from datetime import datetime

import click

from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode

from ..session import Session
from .pretty import print_error


@main.group()
def server_logs():
    """Provides operations related to server logs."""


@server_logs.command()
@click.option(
    "--mark-read", is_flag=True, default=False, help="Mark read flag for server logs being fetched."
)
@click.option(
    "-l", "--page-size", type=int, default=20, help="Number of logs to fetch (from latest log)"
)
@click.option("-n", "--page-number", type=int, default=1, help="Page number to fetch.")
def list(mark_read, page_size, page_number):
    """Fetch server (error) logs."""
    with Session() as session:
        try:
            resp = session.ServerLog.list(mark_read, page_size, page_number)
            logs = resp.get("logs")
            count = resp.get("count", 0)
            if logs is not None:
                print("Total log count:", count)
                for log in logs:
                    log_time = datetime.utcfromtimestamp(log["created_at"]).strftime(
                        "%Y-%m-%d %H:%M:%S"
                    )
                    print("----")
                    print(log_time, log["severity"].upper(), log["source"], log["user"])
                    print(log["request_status"], log["request_url"])
                    print(log["message"])
                    print(log["traceback"])
            else:
                print("No logs.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)



from __future__ import annotations

import abc
import asyncio
import inspect
import queue
import threading
import warnings
from contextvars import Context, ContextVar, copy_context
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Coroutine,
    Iterator,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import aiohttp
from multidict import CIMultiDict

from .config import MIN_API_VERSION, APIConfig, get_config, parse_api_version
from .exceptions import APIVersionWarning, BackendAPIError, BackendClientError
from .types import Sentinel, sentinel

__all__ = (
    "BaseSession",
    "Session",
    "AsyncSession",
    "api_session",
)

from contextlib import asynccontextmanager as actxmgr

from ..common.types import SSLContextType

api_session: ContextVar[BaseSession] = ContextVar("api_session")


async def _negotiate_api_version(
    http_session: aiohttp.ClientSession,
    config: APIConfig,
) -> Tuple[int, str]:
    client_version = parse_api_version(config.version)
    try:
        timeout_config = aiohttp.ClientTimeout(
            total=None,
            connect=None,
            sock_connect=config.connection_timeout,
            sock_read=config.read_timeout,
        )
        headers = CIMultiDict([
            ("User-Agent", config.user_agent),
        ])
        probe_url = (
            config.endpoint / "func/" if config.endpoint_type == "session" else config.endpoint
        )
        async with http_session.get(probe_url, timeout=timeout_config, headers=headers) as resp:
            resp.raise_for_status()
            server_info = await resp.json()
            server_version = parse_api_version(server_info["version"])
            if server_version > client_version:
                warnings.warn(
                    "The server API version is higher than the client. "
                    "Please upgrade the client package.",
                    category=APIVersionWarning,
                )
            if server_version < MIN_API_VERSION:
                warnings.warn(
                    f"The server is too old ({server_version}) and does not meet the minimum API version"
                    f" requirement: v{MIN_API_VERSION[0]}.{MIN_API_VERSION[1]}\nPlease upgrade"
                    " the server or downgrade/reinstall the client SDK with the same"
                    " major.minor release of the server.",
                    category=APIVersionWarning,
                )
            return min(server_version, client_version)
    except (asyncio.TimeoutError, aiohttp.ClientError):
        # fallback to the configured API version
        return client_version


async def _close_aiohttp_session(session: aiohttp.ClientSession) -> None:
    # This is a hacky workaround for premature closing of SSL transports
    # on Windows Proactor event loops.
    # Thanks to Vadim Markovtsev's comment on the aiohttp issue #1925.
    # (https://github.com/aio-libs/aiohttp/issues/1925#issuecomment-592596034)
    transports = 0
    all_is_lost = asyncio.Event()
    if session.connector is None:
        all_is_lost.set()
    else:
        if len(session.connector._conns) == 0:
            all_is_lost.set()
        for conn in session.connector._conns.values():
            for handler, _ in conn:
                proto = getattr(handler.transport, "_ssl_protocol", None)
                if proto is None:
                    continue
                transports += 1
                orig_lost = proto.connection_lost
                orig_eof_received = proto.eof_received

                def connection_lost(exc):
                    orig_lost(exc)
                    nonlocal transports
                    transports -= 1
                    if transports == 0:
                        all_is_lost.set()

                def eof_received():
                    try:
                        orig_eof_received()
                    except AttributeError:
                        # It may happen that eof_received() is called after
                        # _app_protocol and _transport are set to None.
                        pass

                proto.connection_lost = connection_lost
                proto.eof_received = eof_received
    await session.close()
    if transports > 0:
        await all_is_lost.wait()


_Item = TypeVar("_Item")


class _SyncWorkerThread(threading.Thread):
    work_queue: queue.Queue[
        Union[
            Tuple[Union[AsyncIterator, Coroutine], Context],
            Sentinel,
        ]
    ]
    done_queue: queue.Queue[Union[Any, Exception]]
    stream_queue: queue.Queue[Union[Any, Exception, Sentinel]]
    stream_block: threading.Event
    agen_shutdown: bool

    __slots__ = (
        "work_queue",
        "done_queue",
        "stream_queue",
        "stream_block",
        "agen_shutdown",
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.work_queue = queue.Queue()
        self.done_queue = queue.Queue()
        self.stream_queue = queue.Queue()
        self.stream_block = threading.Event()
        self.agen_shutdown = False

    def run(self) -> None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            while True:
                item = self.work_queue.get()
                if item is sentinel:
                    break
                coro, ctx = item
                if inspect.isasyncgen(coro):
                    ctx.run(loop.run_until_complete, self.agen_wrapper(coro))
                else:
                    try:
                        # FIXME: Once python/mypy#12756 is resolved, remove the type-ignore tag.
                        result = ctx.run(loop.run_until_complete, coro)  # type: ignore
                    except Exception as e:
                        self.done_queue.put_nowait(e)
                    else:
                        self.done_queue.put_nowait(result)
                self.work_queue.task_done()
        except (SystemExit, KeyboardInterrupt):
            pass
        finally:
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.stop()
            loop.close()

    def execute(self, coro: Coroutine) -> Any:
        ctx = copy_context()  # preserve context for the worker thread
        try:
            self.work_queue.put((coro, ctx))
            result = self.done_queue.get()
            self.done_queue.task_done()
            if isinstance(result, Exception):
                raise result
            return result
        finally:
            del ctx

    async def agen_wrapper(self, agen):
        self.agen_shutdown = False
        try:
            async for item in agen:
                self.stream_block.clear()
                self.stream_queue.put(item)
                # flow-control the generator.
                self.stream_block.wait()
                if self.agen_shutdown:
                    break
        except Exception as e:
            self.stream_queue.put(e)
        finally:
            self.stream_queue.put(sentinel)
            await agen.aclose()

    def execute_generator(self, asyncgen: AsyncIterator[_Item]) -> Iterator[_Item]:
        ctx = copy_context()  # preserve context for the worker thread
        try:
            self.work_queue.put((asyncgen, ctx))
            while True:
                item = self.stream_queue.get()
                try:
                    if item is sentinel:
                        break
                    if isinstance(item, Exception):
                        raise item
                    yield item
                finally:
                    self.stream_block.set()
                    self.stream_queue.task_done()
        finally:
            del ctx

    def interrupt_generator(self):
        self.agen_shutdown = True
        self.stream_block.set()
        self.stream_queue.put(sentinel)


class BaseSession(metaclass=abc.ABCMeta):
    """
    The base abstract class for sessions.
    """

    __slots__ = (
        "_config",
        "_closed",
        "_context_token",
        "_proxy_mode",
        "aiohttp_session",
        "api_version",
        "System",
        "Manager",
        "Admin",
        "Agent",
        "ContainerRegistry",
        "AgentWatcher",
        "ScalingGroup",
        "Storage",
        "Image",
        "ComputeSession",
        "SessionTemplate",
        "Domain",
        "Group",
        "Auth",
        "User",
        "KeyPair",
        "BackgroundTask",
        "EtcdConfig",
        "Resource",
        "KeypairResourcePolicy",
        "VFolder",
        "Dotfile",
        "ServerLog",
        "Permission",
        "Service",
        "ServiceAutoScalingRule",
        "Model",
        "QuotaScope",
        "Network",
        "UserResourcePolicy",
    )

    aiohttp_session: aiohttp.ClientSession
    api_version: Tuple[int, str]

    _closed: bool
    _config: APIConfig
    _proxy_mode: bool

    def __init__(
        self,
        *,
        config: Optional[APIConfig] = None,
        proxy_mode: bool = False,
    ) -> None:
        self._closed = False
        self._config = config if config else get_config()
        self._proxy_mode = proxy_mode
        self.api_version = parse_api_version(self._config.version)

        from .func.acl import Permission
        from .func.admin import Admin
        from .func.agent import Agent, AgentWatcher
        from .func.auth import Auth
        from .func.bgtask import BackgroundTask
        from .func.container_registry import ContainerRegistry
        from .func.domain import Domain
        from .func.dotfile import Dotfile
        from .func.etcd import EtcdConfig
        from .func.group import Group
        from .func.image import Image
        from .func.keypair import KeyPair
        from .func.keypair_resource_policy import KeypairResourcePolicy
        from .func.manager import Manager
        from .func.model import Model
        from .func.network import Network
        from .func.quota_scope import QuotaScope
        from .func.resource import Resource
        from .func.scaling_group import ScalingGroup
        from .func.server_log import ServerLog
        from .func.service import Service
        from .func.service_auto_scaling_rule import ServiceAutoScalingRule
        from .func.session import ComputeSession
        from .func.session_template import SessionTemplate
        from .func.storage import Storage
        from .func.system import System
        from .func.user import User
        from .func.user_resource_policy import UserResourcePolicy
        from .func.vfolder import VFolderByName

        self.System = System
        self.Admin = Admin
        self.Agent = Agent
        self.AgentWatcher = AgentWatcher
        self.Storage = Storage
        self.Auth = Auth
        self.BackgroundTask = BackgroundTask
        self.ContainerRegistry = ContainerRegistry
        self.EtcdConfig = EtcdConfig
        self.Domain = Domain
        self.Group = Group
        self.Image = Image
        self.ComputeSession = ComputeSession
        self.KeyPair = KeyPair
        self.Manager = Manager
        self.Resource = Resource
        self.KeypairResourcePolicy = KeypairResourcePolicy
        self.User = User
        self.ScalingGroup = ScalingGroup
        self.SessionTemplate = SessionTemplate
        self.VFolder = VFolderByName
        self.Dotfile = Dotfile
        self.ServerLog = ServerLog
        self.Permission = Permission
        self.Service = Service
        self.ServiceAutoScalingRule = ServiceAutoScalingRule
        self.Model = Model
        self.QuotaScope = QuotaScope
        self.Network = Network
        self.UserResourcePolicy = UserResourcePolicy

    @property
    def proxy_mode(self) -> bool:
        """
        If set True, it skips API version negotiation when opening the session.
        """
        return self._proxy_mode

    @abc.abstractmethod
    def open(self) -> Union[None, Awaitable[None]]:
        """
        Initializes the session and perform version negotiation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> Union[None, Awaitable[None]]:
        """
        Terminates the session and releases underlying resources.
        """
        raise NotImplementedError

    @property
    def closed(self) -> bool:
        """
        Checks if the session is closed.
        """
        return self._closed

    @property
    def config(self) -> APIConfig:
        """
        The configuration used by this session object.
        """
        return self._config

    def __enter__(self) -> BaseSession:
        raise NotImplementedError

    def __exit__(self, *exc_info) -> Literal[False]:
        return False

    async def __aenter__(self) -> BaseSession:
        raise NotImplementedError

    async def __aexit__(self, *exc_info) -> Literal[False]:
        return False


class Session(BaseSession):
    """
    A context manager for API client sessions that makes API requests synchronously.
    You may call simple request-response APIs like a plain Python function,
    but cannot use streaming APIs based on WebSocket and Server-Sent Events.
    """

    __slots__ = ("_worker_thread",)

    def __init__(
        self,
        *,
        config: Optional[APIConfig] = None,
        proxy_mode: bool = False,
    ) -> None:
        super().__init__(config=config, proxy_mode=proxy_mode)
        self._worker_thread = _SyncWorkerThread()
        self._worker_thread.start()

        async def _create_aiohttp_session() -> aiohttp.ClientSession:
            ssl: SSLContextType = True
            if self._config.skip_sslcert_validation:
                ssl = False
            connector = aiohttp.TCPConnector(ssl=ssl)
            return aiohttp.ClientSession(connector=connector)

        self.aiohttp_session = self.worker_thread.execute(_create_aiohttp_session())

    def open(self) -> None:
        self._context_token = api_session.set(self)
        if not self._proxy_mode:
            self.api_version = self.worker_thread.execute(
                _negotiate_api_version(self.aiohttp_session, self.config)
            )

    def close(self) -> None:
        """
        Terminates the session.  It schedules the ``close()`` coroutine
        of the underlying aiohttp session and then enqueues a sentinel
        object to indicate termination.  Then it waits until the worker
        thread to self-terminate by joining.
        """
        if self._closed:
            return
        self._closed = True
        self._worker_thread.interrupt_generator()
        self._worker_thread.execute(_close_aiohttp_session(self.aiohttp_session))
        self._worker_thread.work_queue.put(sentinel)
        self._worker_thread.join()
        api_session.reset(self._context_token)

    @property
    def worker_thread(self):
        """
        The thread that internally executes the asynchronous implementations
        of the given API functions.
        """
        return self._worker_thread

    def __enter__(self) -> Session:
        assert not self.closed, "Cannot reuse closed session"
        self.open()
        if self.config.announcement_handler:
            try:
                payload = self.Manager.get_announcement()
                if payload["enabled"]:
                    self.config.announcement_handler(payload["message"])
            except (BackendClientError, BackendAPIError):
                # The server may be an old one without announcement API.
                pass
        return self

    def __exit__(self, *exc_info) -> Literal[False]:
        self.close()
        return False  # raise up the inner exception


def _default_http_client_session(skip_sslcert_validation: bool) -> aiohttp.ClientSession:
    """
    Returns a default aiohttp client session with the default configuration.
    This is used for the API client session when no explicit session is provided.
    """
    ssl: SSLContextType = True
    if skip_sslcert_validation:
        ssl = False
    connector = aiohttp.TCPConnector(ssl=ssl)
    return aiohttp.ClientSession(connector=connector)


class AsyncSession(BaseSession):
    """
    A context manager for API client sessions that makes API requests asynchronously.
    You may call all APIs as coroutines.
    WebSocket-based APIs and SSE-based APIs returns special response types.
    """

    def __init__(
        self,
        *,
        config: Optional[APIConfig] = None,
        proxy_mode: bool = False,
        aiohttp_session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        super().__init__(config=config, proxy_mode=proxy_mode)
        if aiohttp_session is not None:
            self.aiohttp_session = aiohttp_session
            self._aiohttp_session_injected = True
        else:
            self.aiohttp_session = _default_http_client_session(
                self._config.skip_sslcert_validation
            )
            self._aiohttp_session_injected = False

    async def _aopen(self) -> None:
        self._context_token = api_session.set(self)
        if not self._proxy_mode:
            self.api_version = await _negotiate_api_version(self.aiohttp_session, self.config)

    def open(self) -> Awaitable[None]:
        return self._aopen()

    async def _aclose(self) -> None:
        if self._closed:
            return
        self._closed = True
        if not self._aiohttp_session_injected:
            await _close_aiohttp_session(self.aiohttp_session)
        api_session.reset(self._context_token)

    def close(self) -> Awaitable[None]:
        return self._aclose()

    async def __aenter__(self) -> AsyncSession:
        assert not self.closed, "Cannot reuse closed session"
        await self.open()
        if self.config.announcement_handler:
            try:
                payload = await self.Manager.get_announcement()
                if payload["enabled"]:
                    self.config.announcement_handler(payload["message"])
            except (BackendClientError, BackendAPIError):
                # The server may be an old one without announcement API.
                pass
        return self

    async def __aexit__(self, *exc_info) -> Literal[False]:
        await self.close()
        return False  # raise up the inner exception


# TODO: Remove this after refactoring session management with contextvars
@actxmgr
async def set_api_context(session):
    token = api_session.set(session)
    try:
        yield
    finally:
        api_session.reset(token)



from __future__ import annotations

import enum
from typing import Any, Callable, MutableMapping

from ai.backend.cli.types import Undefined, undefined

__all__ = (
    "Sentinel",
    "sentinel",
)


def set_if_set(
    target_dict: MutableMapping[str, Any],
    key: str,
    value: Any | Undefined,
    *,
    clean_func: Callable[[Any], Any] | None = None,
) -> None:
    if value is not undefined:
        if clean_func is not None:
            value = clean_func(value)
        target_dict[key] = value


class Sentinel(enum.Enum):
    """
    A special type to represent a special value to indicate closing/shutdown of queues.
    """

    TOKEN = 0

    def __bool__(self) -> bool:
        # It should be evaluated as False when used as a boolean expr.
        return False


sentinel = Sentinel.TOKEN



from typing import Mapping, Sequence, Union

from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("ServerLog",)


class ServerLog(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches various server logs.
    """

    @api_function
    @classmethod
    async def list(
        cls,
        mark_read: bool = False,
        page_size: int = 20,
        page_no: int = 1,
    ) -> Sequence[dict]:
        """
        Fetches server (error) logs.

        :param mark_read: Mark read flog for server logs being fetched.
        :param page_size: Number of logs to fetch (from latest log).
        :param page_no: Page number to fetch.
        """
        params: Mapping[str, Union[str, int]] = {
            "mark_read": str(mark_read),
            "page_size": page_size,
            "page_no": page_no,
        }
        rqst = Request("GET", "/logs/error", params=params)
        async with rqst.fetch() as resp:
            return await resp.json()



from __future__ import annotations

import asyncio
import functools
import io
import json as modjson
import logging
import sys
import uuid
from collections import namedtuple
from datetime import datetime
from decimal import Decimal
from enum import StrEnum
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import aiohttp
import aiohttp.web
import appdirs
import attrs
from aiohttp.client import _RequestContextManager, _WSRequestContextManager
from dateutil.tz import tzutc
from multidict import CIMultiDict
from yarl import URL

from ai.backend.common.json import dump_json_str

from .auth import generate_signature
from .exceptions import BackendAPIError, BackendClientError
from .session import AsyncSession, BaseSession, api_session
from .session import Session as SyncSession

log = logging.getLogger(__spec__.name)

__all__ = [
    "Request",
    "BaseResponse",
    "Response",
    "WebSocketResponse",
    "SSEResponse",
    "FetchContextManager",
    "WebSocketContextManager",
    "SSEContextManager",
    "AttachedFile",
]


RequestContent = Union[
    bytes,
    bytearray,
    str,
    aiohttp.StreamReader,
    io.IOBase,
    None,
]
"""
The type alias for the set of allowed types for request content.
"""


AttachedFile = namedtuple("AttachedFile", "filename stream content_type")
"""
A struct that represents an attached file to the API request.

:param str filename: The name of file to store. It may include paths
                     and the server will create parent directories
                     if required.

:param Any stream: A file-like object that allows stream-reading bytes.

:param str content_type: The content type for the stream.  For arbitrary
                         binary data, use "application/octet-stream".
"""


_T = TypeVar("_T")


async def _coro_return(val: _T) -> _T:
    return val


class ExtendedJSONEncoder(modjson.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, uuid.UUID):
            return str(obj)
        if isinstance(obj, Path):
            return str(obj)
        if isinstance(obj, Decimal):
            return str(obj)
        return super().default(obj)


class SessionMode(StrEnum):
    CLIENT = "CLIENT"
    PROXY = "PROXY"


class Request:
    """
    The API request object.
    """

    __slots__ = (
        "config",
        "session",
        "method",
        "path",
        "date",
        "headers",
        "params",
        "content_type",
        "api_version",
        "_content",
        "_attached_files",
        "reporthook",
        "_session_mode",
    )

    _content: RequestContent
    _attached_files: Optional[Sequence[AttachedFile]]

    date: Optional[datetime]
    api_version: str

    _allowed_methods = frozenset(["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"])

    def __init__(
        self,
        method: str = "GET",
        path: Optional[str] = None,
        content: Optional[RequestContent] = None,
        *,
        content_type: Optional[str] = None,
        params: Optional[Mapping[str, Union[str, int]]] = None,
        reporthook: Optional[Callable] = None,
        override_api_version: Optional[str] = None,
        session_mode: SessionMode = SessionMode.CLIENT,
    ) -> None:
        """
        Initialize an API request.

        :param BaseSession session: The session where this request is executed on.

        :param str path: The query path. When performing requests, the version number
                         prefix will be automatically prepended if required.

        :param RequestContent content: The API query body which will be encoded as
                                       JSON.

        :param str content_type: Explicitly set the content type.  See also
                                 :func:`Request.set_content`.

        :param SessionMode session_mode: The session mode, either CLIENT or PROXY.
                                      This affects how the response decoding is handled.
        """
        self.session = api_session.get()
        self.config = self.session.config
        self.method = method
        if path is not None and path.startswith("/"):
            path = path[1:]
        self.path = path
        self.params = params
        self.date = None
        if override_api_version:
            self.api_version = override_api_version
        else:
            self.api_version = f"v{self.session.api_version[0]}.{self.session.api_version[1]}"
        self.headers = CIMultiDict([
            ("User-Agent", self.config.user_agent),
            ("X-BackendAI-Domain", self.config.domain),
            ("X-BackendAI-Version", self.api_version),
        ])
        self._content = b""
        self._attached_files = None
        self.set_content(content, content_type=content_type)
        self.reporthook = reporthook
        self._session_mode = session_mode

    @property
    def content(self) -> RequestContent:
        """
        Retrieves the content in the original form.
        Private codes should NOT use this as it incurs duplicate
        encoding/decoding.
        """
        return self._content

    def set_content(
        self,
        value: RequestContent,
        *,
        content_type: Optional[str] = None,
    ) -> None:
        """
        Sets the content of the request.
        """
        assert self._attached_files is None, (
            "cannot set content because you already attached files."
        )
        guessed_content_type = "application/octet-stream"
        if value is None:
            guessed_content_type = "text/plain"
            self._content = b""
        elif isinstance(value, str):
            guessed_content_type = "text/plain"
            self._content = value.encode("utf-8")
        else:
            guessed_content_type = "application/octet-stream"
            self._content = value
        self.content_type = content_type if content_type is not None else guessed_content_type

    def set_json(self, value: Any) -> None:
        """
        A shortcut for set_content() with JSON objects.
        """
        self.set_content(
            modjson.dumps(value, cls=ExtendedJSONEncoder), content_type="application/json"
        )

    def attach_files(self, files: Sequence[AttachedFile]) -> None:
        """
        Attach a list of files represented as AttachedFile.
        """
        assert not self._content, "content must be empty to attach files."
        self.content_type = "multipart/form-data"
        self._attached_files = files

    def _sign(
        self,
        rel_url: URL,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        hash_type: Optional[str] = None,
    ) -> None:
        """
        Calculates the signature of the given request and adds the
        Authorization HTTP header.
        It should be called at the very end of request preparation and before
        sending the request to the server.
        """
        if access_key is None:
            access_key = self.config.access_key
        if secret_key is None:
            secret_key = self.config.secret_key
        if hash_type is None:
            hash_type = self.config.hash_type
        assert self.date is not None
        if self.config.endpoint_type == "api":
            hdrs, _ = generate_signature(
                method=self.method,
                version=self.api_version,
                endpoint=self.config.endpoint,
                date=self.date,
                rel_url=str(rel_url),
                content_type=self.content_type,
                access_key=access_key,
                secret_key=secret_key,
                hash_type=hash_type,
            )
            self.headers.update(hdrs)
        elif self.config.endpoint_type == "session":
            local_state_path = Path(appdirs.user_state_dir("backend.ai", "Lablup"))
            try:
                cookie_jar = cast(aiohttp.CookieJar, self.session.aiohttp_session.cookie_jar)
                cookie_jar.load(local_state_path / "cookie.dat")
            except (IOError, PermissionError):
                pass
        else:
            raise ValueError("unsupported endpoint type")

    def _pack_content(self) -> Union[RequestContent, aiohttp.FormData]:
        if self._attached_files is not None:
            data = aiohttp.FormData()
            for f in self._attached_files:
                data.add_field("src", f.stream, filename=f.filename, content_type=f.content_type)
            assert data.is_multipart, "Failed to pack files as multipart."
            # Let aiohttp fill up the content-type header including
            # multipart boundaries.
            self.headers.pop("Content-Type", None)
            return data
        else:
            return self._content

    def _build_url(self) -> URL:
        base_url = self.config.endpoint.path.rstrip("/")
        query_path = self.path.lstrip("/") if self.path is not None and len(self.path) > 0 else ""
        if self.config.endpoint_type == "session":
            if not query_path.startswith("server"):
                query_path = "func/{0}".format(query_path)
        path = "{0}/{1}".format(base_url, query_path)
        url = self.config.endpoint.with_path(path)
        if self.params:
            url = url.with_query(self.params)
        return url

    # TODO: attach rate-limit information

    def fetch(self, **kwargs) -> FetchContextManager:
        """
        Sends the request to the server and reads the response.

        You may use this method with AsyncSession only,
        following the pattern below:

        .. code-block:: python3

          from ai.backend.client.request import Request
          from ai.backend.client.session import AsyncSession

          async with AsyncSession() as sess:
            rqst = Request('GET', ...)
            async with rqst.fetch() as resp:
              print(await resp.text())
        """
        assert self.method in self._allowed_methods, "Disallowed HTTP method: {}".format(
            self.method
        )
        self.date = datetime.now(tzutc())
        assert self.date is not None
        self.headers["Date"] = self.date.isoformat()
        if self.content_type is not None and "Content-Type" not in self.headers:
            self.headers["Content-Type"] = self.content_type
        force_anonymous = kwargs.pop("anonymous", False)

        def _rqst_ctx_builder():
            timeout_config = aiohttp.ClientTimeout(
                total=None,
                connect=None,
                sock_connect=self.config.connection_timeout,
                sock_read=self.config.read_timeout,
            )
            full_url = self._build_url()
            if self.session.aiohttp_session._base_url is None:
                # for anonymous requests or API sessions created without using ClientPool
                request_url = full_url
            else:
                request_url = full_url.relative()
            if not self.config.is_anonymous and not force_anonymous:
                self._sign(full_url.relative())
            return self.session.aiohttp_session.request(
                self.method,
                str(request_url),
                data=self._pack_content(),
                timeout=timeout_config,
                headers=self.headers,
                allow_redirects=False,
            )

        return FetchContextManager(self.session, _rqst_ctx_builder, self._session_mode, **kwargs)

    def connect_websocket(self, **kwargs) -> WebSocketContextManager:
        """
        Creates a WebSocket connection.

        .. warning::

          This method only works with
          :class:`~ai.backend.client.session.AsyncSession`.
        """
        assert isinstance(self.session, AsyncSession), (
            "Cannot use websockets with sessions in the synchronous mode"
        )
        assert self.method == "GET", "Invalid websocket method"
        self.date = datetime.now(tzutc())
        assert self.date is not None
        self.headers["Date"] = self.date.isoformat()
        # websocket is always a "binary" stream.
        self.content_type = "application/octet-stream"

        def _ws_ctx_builder():
            full_url = self._build_url()
            if not self.config.is_anonymous:
                self._sign(full_url.relative())
            if self.session.aiohttp_session._base_url is None:
                # for anonymous requests or API sessions created without using ClientPool
                request_url = full_url
            else:
                request_url = full_url.relative()
            return self.session.aiohttp_session.ws_connect(
                str(request_url), autoping=True, heartbeat=30.0, headers=self.headers
            )

        return WebSocketContextManager(self.session, _ws_ctx_builder, **kwargs)

    def connect_events(self, **kwargs) -> SSEContextManager:
        """
        Creates a Server-Sent Events connection.

        .. warning::

          This method only works with
          :class:`~ai.backend.client.session.AsyncSession`.
        """
        assert isinstance(self.session, AsyncSession), (
            "Cannot use event streams with sessions in the synchronous mode"
        )
        assert self.method == "GET", "Invalid event stream method"
        self.date = datetime.now(tzutc())
        assert self.date is not None
        self.headers["Date"] = self.date.isoformat()
        self.content_type = "application/octet-stream"

        def _rqst_ctx_builder():
            timeout_config = aiohttp.ClientTimeout(
                total=None,
                connect=None,
                sock_connect=self.config.connection_timeout,
                sock_read=self.config.read_timeout,
            )
            full_url = self._build_url()
            if not self.config.is_anonymous:
                self._sign(full_url.relative())
            if self.session.aiohttp_session._base_url is None:
                # for anonymous requests or API sessions created without using ClientPool
                request_url = full_url
            else:
                request_url = full_url.relative()
            return self.session.aiohttp_session.request(
                self.method, str(request_url), timeout=timeout_config, headers=self.headers
            )

        return SSEContextManager(self.session, _rqst_ctx_builder, **kwargs)


class AsyncResponseMixin:
    _session: BaseSession
    _raw_response: aiohttp.ClientResponse

    async def text(self) -> str:
        return await self._raw_response.text()

    async def json(self, *, loads=modjson.loads) -> Any:
        loads = functools.partial(loads)
        return await self._raw_response.json(loads=loads)

    async def read(self, n: int = -1) -> bytes:
        return await self._raw_response.content.read(n)

    async def readall(self) -> bytes:
        return await self._raw_response.content.read(-1)


class SyncResponseMixin:
    _session: BaseSession
    _raw_response: aiohttp.ClientResponse

    def text(self) -> str:
        sync_session = cast(SyncSession, self._session)
        return sync_session.worker_thread.execute(
            self._raw_response.text(),
        )

    def json(self, *, loads=modjson.loads) -> Any:
        loads = functools.partial(loads)
        sync_session = cast(SyncSession, self._session)
        return sync_session.worker_thread.execute(
            self._raw_response.json(loads=loads),
        )

    def read(self, n: int = -1) -> bytes:
        sync_session = cast(SyncSession, self._session)
        return sync_session.worker_thread.execute(
            self._raw_response.content.read(n),
        )

    def readall(self) -> bytes:
        sync_session = cast(SyncSession, self._session)
        return sync_session.worker_thread.execute(
            self._raw_response.content.read(-1),
        )


class BaseResponse:
    """
    Represents the Backend.AI API response.
    Also serves as a high-level wrapper of :class:`aiohttp.ClientResponse`.

    The response objects are meant to be created by the SDK, not the callers.

    :func:`text`, :func:`json` methods return the resolved content directly with
    plain synchronous Session while they return the coroutines with AsyncSession.
    """

    __slots__ = (
        "_session",
        "_raw_response",
        "_async_mode",
    )

    _session: BaseSession
    _raw_response: aiohttp.ClientResponse
    _async_mode: bool

    def __init__(
        self,
        session: BaseSession,
        underlying_response: aiohttp.ClientResponse,
        *,
        async_mode: bool = False,
        **kwargs,
    ) -> None:
        self._session = session
        self._raw_response = underlying_response
        self._async_mode = async_mode

    @property
    def session(self) -> BaseSession:
        return self._session

    @property
    def status(self) -> int:
        return self._raw_response.status

    @property
    def reason(self) -> str:
        if self._raw_response.reason is not None:
            return self._raw_response.reason
        return ""

    @property
    def headers(self) -> Mapping[str, str]:
        return self._raw_response.headers

    @property
    def raw_response(self) -> aiohttp.ClientResponse:
        return self._raw_response

    @property
    def content_type(self) -> str:
        return self._raw_response.content_type

    @property
    def content_length(self) -> Optional[int]:
        return self._raw_response.content_length

    @property
    def content(self) -> aiohttp.StreamReader:
        return self._raw_response.content


class Response(AsyncResponseMixin, BaseResponse):
    pass


class FetchContextManager:
    """
    The context manager returned by :func:`Request.fetch`.

    It provides asynchronous context manager interfaces only.
    """

    __slots__ = (
        "session",
        "rqst_ctx_builder",
        "_session_mode",
        "response_cls",
        "check_status",
        "_async_mode",
        "_rqst_ctx",
    )

    _rqst_ctx: Optional[_RequestContextManager]

    def __init__(
        self,
        session: BaseSession,
        rqst_ctx_builder: Callable[[], _RequestContextManager],
        session_mode: SessionMode = SessionMode.CLIENT,
        *,
        response_cls: Type[Response] = Response,
        check_status: bool = True,
    ) -> None:
        self.session = session
        self.rqst_ctx_builder = rqst_ctx_builder
        self.check_status = check_status
        self.response_cls = response_cls
        self._async_mode = isinstance(session, AsyncSession)
        self._rqst_ctx = None
        self._session_mode = session_mode

    async def __aenter__(self) -> Response:
        max_retries = len(self.session.config.endpoints)
        retry_count = 0
        raw_resp: Optional[aiohttp.ClientResponse] = None
        while True:
            try:
                retry_count += 1
                self._rqst_ctx = self.rqst_ctx_builder()
                assert self._rqst_ctx is not None
                raw_resp = await self._rqst_ctx.__aenter__()
                if self.check_status and raw_resp.status // 100 not in [2, 3]:
                    match self._session_mode:
                        case SessionMode.CLIENT:
                            error_data = await raw_resp.json()
                            msg = dump_json_str(error_data)
                            await raw_resp.__aexit__(None, None, None)
                            raise BackendAPIError(raw_resp.status, raw_resp.reason or "", msg)
                        case SessionMode.PROXY:
                            pass
                return self.response_cls(self.session, raw_resp, async_mode=self._async_mode)
            except aiohttp.ClientConnectionError as e:
                if retry_count == max_retries:
                    msg = (
                        "Request to the API endpoint has failed.\n"
                        "Check your network connection and/or the server status.\n"
                        "\u279c {!r}".format(e)
                    )
                    raise BackendClientError(msg) from e
                else:
                    self.session.config.rotate_endpoints()
                    continue
            except aiohttp.ClientResponseError as e:
                msg = "API endpoint response error.\n\u279c {!r}".format(e)
                if raw_resp is not None:
                    await raw_resp.__aexit__(*sys.exc_info())
                raise BackendClientError(msg) from e
            finally:
                self.session.config.load_balance_endpoints()

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        assert self._rqst_ctx is not None
        await self._rqst_ctx.__aexit__(*exc_info)
        self._rqst_ctx = None
        return None


class WebSocketResponse(BaseResponse):
    """
    A high-level wrapper of :class:`aiohttp.ClientWebSocketResponse`.
    """

    __slots__ = ("_raw_ws",)

    def __init__(
        self,
        session: BaseSession,
        underlying_response: aiohttp.ClientResponse,
        **kwargs,
    ) -> None:
        # Unfortunately, aiohttp.ClientWebSocketResponse is not a subclass of aiohttp.ClientResponse.
        # Since we block methods that require ClientResponse-specific methods, we just force-typecast.
        super().__init__(session, underlying_response, **kwargs)
        self._raw_ws = cast(aiohttp.ClientWebSocketResponse, underlying_response)

    @property
    def content_type(self) -> str:
        raise AttributeError("WebSocketResponse does not have an explicit content type.")

    @property
    def content_length(self) -> Optional[int]:
        raise AttributeError("WebSocketResponse does not have a fixed content length.")

    @property
    def content(self) -> aiohttp.StreamReader:
        raise AttributeError("WebSocketResponse does not support reading the content.")

    @property
    def raw_websocket(self) -> aiohttp.ClientWebSocketResponse:
        return self._raw_ws

    @property
    def closed(self) -> bool:
        return self._raw_ws.closed

    async def close(self) -> None:
        await self._raw_ws.close()

    def __aiter__(self) -> AsyncIterator[aiohttp.WSMessage]:
        return self._raw_ws.__aiter__()

    def exception(self) -> Optional[BaseException]:
        return self._raw_ws.exception()

    async def send_str(self, raw_str: str) -> None:
        if self._raw_ws.closed:
            raise aiohttp.ServerDisconnectedError("server disconnected")
        await self._raw_ws.send_str(raw_str)

    async def send_json(self, obj: Any) -> None:
        if self._raw_ws.closed:
            raise aiohttp.ServerDisconnectedError("server disconnected")
        await self._raw_ws.send_json(obj)

    async def send_bytes(self, data: bytes) -> None:
        if self._raw_ws.closed:
            raise aiohttp.ServerDisconnectedError("server disconnected")
        await self._raw_ws.send_bytes(data)

    async def receive_str(self) -> str:
        if self._raw_ws.closed:
            raise aiohttp.ServerDisconnectedError("server disconnected")
        return await self._raw_ws.receive_str()

    async def receive_json(self) -> Any:
        if self._raw_ws.closed:
            raise aiohttp.ServerDisconnectedError("server disconnected")
        return await self._raw_ws.receive_json()

    async def receive_bytes(self) -> bytes:
        if self._raw_ws.closed:
            raise aiohttp.ServerDisconnectedError("server disconnected")
        return await self._raw_ws.receive_bytes()


class WebSocketContextManager:
    """
    The context manager returned by :func:`Request.connect_websocket`.
    """

    __slots__ = (
        "session",
        "ws_ctx_builder",
        "response_cls",
        "on_enter",
        "_ws_ctx",
    )

    _ws_ctx: Optional[_WSRequestContextManager]

    def __init__(
        self,
        session: BaseSession,
        ws_ctx_builder: Callable[[], _WSRequestContextManager],
        *,
        on_enter: Optional[Callable] = None,
        response_cls: Type[WebSocketResponse] = WebSocketResponse,
    ) -> None:
        self.session = session
        self.ws_ctx_builder = ws_ctx_builder
        self.response_cls = response_cls
        self.on_enter = on_enter
        self._ws_ctx = None

    async def __aenter__(self) -> WebSocketResponse:
        max_retries = len(self.session.config.endpoints)
        retry_count = 0
        while True:
            try:
                retry_count += 1
                self._ws_ctx = self.ws_ctx_builder()
                assert self._ws_ctx is not None
                raw_ws = await self._ws_ctx.__aenter__()
            except aiohttp.ClientConnectionError as e:
                if retry_count == max_retries:
                    msg = (
                        "Request to the API endpoint has failed.\n"
                        "Check your network connection and/or the server status.\n"
                        "Error detail: {!r}".format(e)
                    )
                    raise BackendClientError(msg) from e
                else:
                    self.session.config.rotate_endpoints()
                    continue
            except aiohttp.ClientResponseError as e:
                msg = "API endpoint response error.\n\u279c {!r}".format(e)
                raise BackendClientError(msg) from e
            else:
                break
            finally:
                self.session.config.load_balance_endpoints()

        wrapped_ws = self.response_cls(self.session, cast(aiohttp.ClientResponse, raw_ws))
        if self.on_enter is not None:
            await self.on_enter(wrapped_ws)
        return wrapped_ws

    async def __aexit__(self, *args) -> Optional[bool]:
        assert self._ws_ctx is not None
        await self._ws_ctx.__aexit__(*args)
        self._ws_ctx = None
        return None


@attrs.define(auto_attribs=True, slots=True, frozen=True)
class SSEMessage:
    event: str
    data: str
    id: Optional[str] = None
    retry: Optional[int] = None


class SSEResponse(BaseResponse):
    __slots__ = (
        "_auto_reconnect",
        "_retry",
        "_connector",
    )

    def __init__(
        self,
        session: BaseSession,
        underlying_response: aiohttp.ClientResponse,
        *,
        connector: Callable[[], Awaitable[aiohttp.ClientResponse]],
        auto_reconnect: bool = True,
        default_retry: int = 5,
        **kwargs,
    ) -> None:
        super().__init__(session, underlying_response, async_mode=True, **kwargs)
        self._auto_reconnect = auto_reconnect
        self._retry = default_retry
        self._connector = connector

    async def fetch_events(self) -> AsyncIterator[SSEMessage]:
        msg_lines: List[str] = []
        server_closed = False
        while True:
            received_line = await self._raw_response.content.readline()
            if not received_line:
                # connection closed
                if self._auto_reconnect and not server_closed:
                    await asyncio.sleep(self._retry)
                    self._raw_response = await self._connector()
                    continue
                else:
                    break
            received_line = received_line.strip(b"\r\n")
            if received_line.startswith(b":"):
                # comment
                continue
            if not received_line:
                # message boundary
                if len(msg_lines) == 0:
                    continue
                event_type = "message"
                event_id = None
                event_retry = None
                data_lines = []
                try:
                    for stored_line in msg_lines:
                        hdr, text = stored_line.split(":", maxsplit=1)
                        text = text.lstrip(" ")
                        if hdr == "data":
                            data_lines.append(text)
                        elif hdr == "event":
                            event_type = text
                        elif hdr == "id":
                            event_id = text
                        elif hdr == "retry":
                            event_retry = int(text)
                except (IndexError, ValueError):
                    log.exception("SSEResponse: parsing-error")
                    continue
                event_data = "\n".join(data_lines)
                msg_lines.clear()
                if event_retry is not None:
                    self._retry = event_retry
                yield SSEMessage(
                    event=event_type,
                    data=event_data,
                    id=event_id,
                    retry=event_retry,
                )
                if event_type == "server_close":
                    server_closed = True
                    break
            else:
                msg_lines.append(received_line.decode("utf-8"))

    def __aiter__(self) -> AsyncIterator[SSEMessage]:
        return self.fetch_events()


class SSEContextManager:
    __slots__ = (
        "session",
        "rqst_ctx_builder",
        "response_cls",
        "_rqst_ctx",
    )

    _rqst_ctx: Optional[_RequestContextManager]

    def __init__(
        self,
        session: BaseSession,
        rqst_ctx_builder: Callable[[], _RequestContextManager],
        *,
        response_cls: Type[SSEResponse] = SSEResponse,
    ) -> None:
        self.session = session
        self.rqst_ctx_builder = rqst_ctx_builder
        self.response_cls = response_cls
        self._rqst_ctx = None

    async def reconnect(self) -> aiohttp.ClientResponse:
        if self._rqst_ctx is not None:
            await self._rqst_ctx.__aexit__(None, None, None)
        self._rqst_ctx = self.rqst_ctx_builder()
        assert self._rqst_ctx is not None
        raw_resp = await self._rqst_ctx.__aenter__()
        if raw_resp.status // 100 != 2:
            msg = await raw_resp.text()
            raise BackendAPIError(raw_resp.status, raw_resp.reason or "", msg)
        return raw_resp

    async def __aenter__(self) -> SSEResponse:
        max_retries = len(self.session.config.endpoints)
        retry_count = 0
        while True:
            try:
                retry_count += 1
                raw_resp = await self.reconnect()
                return self.response_cls(self.session, raw_resp, connector=self.reconnect)
            except aiohttp.ClientConnectionError as e:
                if retry_count == max_retries:
                    msg = (
                        "Request to the API endpoint has failed.\n"
                        "Check your network connection and/or the server status.\n"
                        "\u279c {!r}".format(e)
                    )
                    raise BackendClientError(msg) from e
                else:
                    self.session.config.rotate_endpoints()
                    continue
            except aiohttp.ClientResponseError as e:
                msg = "API endpoint response error.\n\u279c {!r}".format(e)
                raise BackendClientError(msg) from e
            finally:
                self.session.config.load_balance_endpoints()

    async def __aexit__(self, *args) -> Optional[bool]:
        assert self._rqst_ctx is not None
        await self._rqst_ctx.__aexit__(*args)
        self._rqst_ctx = None
        return None



from typing import Any, Mapping, Optional

from ..exceptions import BackendAPIError
from ..request import Request
from ..session import api_session
from .base import BaseFunction, api_function

__all__ = ("Admin",)


class Admin(BaseFunction):
    """
    Provides the function interface for making admin GraphQL queries.

    .. note::

      Depending on the privilege of your API access key, you may or may not
      have access to querying/mutating server-side resources of other
      users.
    """

    @api_function
    @classmethod
    async def query(
        cls,
        query: str,
        variables: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """
        Sends the GraphQL query and returns the response.

        :param query: The GraphQL query string.
        :param variables: An optional key-value dictionary
            to fill the interpolated template variables
            in the query.

        :returns: The object parsed from the response JSON string.
        """
        return await cls._query(query, variables)

    @classmethod
    async def _query(
        cls,
        query: str,
        variables: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """
        Internal async implementation of the query() method,
        which may be reused by other functional APIs to make GQL requests.
        """
        gql_query = {
            "query": query,
            "variables": variables if variables else {},
        }
        if api_session.get().api_version >= (6, "20210815"):
            rqst = Request("POST", "/admin/gql")
            rqst.set_json(gql_query)
            async with rqst.fetch() as resp:
                response = await resp.json()
                errors = response.get("errors", [])
                if errors:
                    raise BackendAPIError(
                        400,
                        reason="Bad request",
                        data={
                            "type": "https://api.backend.ai/probs/graphql-error",
                            "title": "GraphQL-generated error",
                            "data": errors,
                        },
                    )
                else:
                    return response["data"]
        else:
            rqst = Request("POST", "/admin/graphql")
            rqst.set_json(gql_query)
            async with rqst.fetch() as resp:
                return await resp.json()



from __future__ import annotations

import sys
from typing import Any, Callable, Iterator, List, Mapping, Optional, Sequence

from tabulate import tabulate

from ai.backend.client.cli.pagination import echo_via_pager, get_preferred_page_size, tabulate_items
from ai.backend.client.cli.pretty import print_done, print_error, print_fail
from ai.backend.common.types import ResultSet

from .types import BaseOutputHandler, FieldSpec, PaginatedResult

_Item = Mapping[str, Any]


class NoItems(Exception):
    pass


class ConsoleOutputHandler(BaseOutputHandler):
    def print_item(
        self,
        item: Optional[_Item],
        fields: Sequence[FieldSpec],
    ) -> None:
        if item is None:
            print_fail("No matching entry found.")
            return
        field_map = {f.field_name: f for f in fields}
        print(
            tabulate(
                [
                    (
                        field_map[k].humanized_name,
                        field_map[k].formatter.format_console(v, field_map[k]),
                    )
                    for k, v in item.items()
                    if k in field_map
                ],
                headers=("Field", "Value"),
            )
        )

    def print_items(
        self,
        items: Sequence[_Item],
        fields: Sequence[FieldSpec],
    ) -> None:
        field_map = {f.field_name: f for f in fields}
        for idx, item in enumerate(items):
            if idx > 0:
                print("-" * 20)
            print(
                tabulate(
                    [
                        (
                            field_map[k].humanized_name,
                            field_map[k].formatter.format_console(v, field_map[k]),
                        )
                        for k, v in item.items()
                        if k in field_map
                    ],
                    headers=("Field", "Value"),
                )
            )

    def print_result_set(
        self,
        result_set: ResultSet,
    ) -> None:
        if result_set["success"]:
            print_done("Successfully created:")
            print(
                tabulate(
                    map(lambda item: [item["item"]], result_set["success"]),
                    tablefmt="plain",
                )
            )
        if result_set["failed"]:
            print_fail("Failed to create:")
            print(
                tabulate(
                    map(lambda item: [item["item"], item["msg"]], result_set["failed"]),
                    tablefmt="plain",
                )
            )

    def print_list(
        self,
        items: Sequence[_Item],
        fields: Sequence[FieldSpec],
        *,
        is_scalar: bool = False,
    ) -> None:
        if is_scalar:
            assert len(fields) == 1
        if sys.stdout.isatty():

            def infinite_fetch():
                current_offset = 0
                page_size = get_preferred_page_size()
                while True:
                    if len(items) == 0:
                        if not sys.stdout.isatty():
                            break
                        else:
                            raise NoItems
                    if is_scalar:
                        yield from map(
                            lambda v: {fields[0].field_name: v},
                            items[current_offset : current_offset + page_size],
                        )
                    else:
                        yield from items[current_offset : current_offset + page_size]
                    current_offset += page_size
                    if current_offset >= len(items):
                        break

            try:
                echo_via_pager(
                    tabulate_items(
                        infinite_fetch(),
                        fields,
                    ),
                )
            except NoItems:
                print("No matching items.")
        else:
            if is_scalar:
                for line in tabulate_items(
                    map(lambda v: {fields[0].field_name: v}, items),  # type: ignore
                    fields,
                ):
                    print(line, end="")
            else:
                for line in tabulate_items(
                    items,  # type: ignore
                    fields,
                ):
                    print(line, end="")

    def print_paginated_list(
        self,
        fetch_func: Callable[[int, int], PaginatedResult],
        initial_page_offset: int,
        page_size: Optional[int] = None,
        plain=False,
    ) -> None:
        fields: List[FieldSpec] = []

        def infinite_fetch(_page_size: int) -> Iterator[_Item]:
            nonlocal fields
            current_offset = initial_page_offset
            while True:
                result = fetch_func(current_offset, _page_size)
                if not fields:
                    fields.extend(result.fields)

                if result.total_count == 0:
                    if not sys.stdout.isatty():
                        break
                    else:
                        raise NoItems
                current_offset += len(result.items)
                yield from result.items
                if current_offset >= result.total_count:
                    break

        if sys.stdout.isatty() and page_size is None:
            preferred_page_size = get_preferred_page_size()
            try:
                echo_via_pager(
                    tabulate_items(
                        infinite_fetch(preferred_page_size),
                        fields,
                        tablefmt="plain" if plain else "simple",
                    ),
                )
            except NoItems:
                print("No matching items.")
        else:
            if page_size is None:
                page_size = 20
            try:
                for line in tabulate_items(
                    infinite_fetch(page_size),
                    fields,
                    tablefmt="plain" if plain else "simple",
                ):
                    print(line, end="")
            except NoItems:
                print("No matching items.")

    def print_mutation_result(
        self,
        item: _Item,
        item_name: Optional[str] = None,
        action_name: Optional[str] = None,
        extra_info: Mapping = {},
    ) -> None:
        t = [
            ["ok", item["ok"]],
            ["msg", item["msg"]],
            *[(k, v) for k, v in extra_info.items()],
        ]
        if action_name is not None:
            t += [["Action", action_name]]
        if item_name is not None:
            t += [(k, v) for k, v in item[item_name].items()]
        print(
            tabulate(
                t,
                headers=("Field", "Value"),
            )
        )

    def print_mutation_error(
        self,
        error: Optional[Exception] = None,
        msg: str = "Failed",
        item_name: Optional[str] = None,
        action_name: Optional[str] = None,
        extra_info: Mapping = {},
    ) -> None:
        t = [
            ["Message", msg],
        ]
        if item_name is not None:
            t += [["Item", item_name]]
        if action_name is not None:
            t += [["Action", action_name]]
        print(
            tabulate(
                t,
                headers=("Field", "Value"),
            )
        )
        if error is not None:
            print_error(error)

    def print_error(
        self,
        error: Exception,
    ) -> None:
        print_error(error)

    def print_fail(
        self,
        message: str,
    ) -> None:
        print_fail(message)
